#ifndef _MAGNA_GRAPH_H
#define _MAGNA_GRAPH_H

#include <stdio.h>
#include "permutation.h"
#include "utils.h"

#include <algorithm> // std::min

struct spmat {
    int nr,nc; // num. rows, cols
    int nnz;
    int *IA; // ind first elt of each row. nnz = IA[nr]
    int *JA; // col inds
    int *INZ; // nnz of each row
    struct dymatelt *A; // number of events
    //int *Aalloc; // number of events allocated
    //float **Astart; // values, start times
    //float **Aend; // values, end times
};

void print_spmat(struct spmat sp);
struct spmat spmat_malloc(int nr, int nc, int nnz);
struct spmat edge_set_to_spmat(struct edge **edge_set, int nk, int nnz, char *filename);

int merge_overlapping(int n, float* tstart, float* tend);
void conserved_duration(int n, float* tstart, float* tend, float *Tc, float *Tn);
void conserved_duration(int nl, float* tstartl, float* tendl,
                        int nr, float* tstartr, float* tendr,
                        float *Tc, float *Tn);

struct vertex {
	char* name;
	int number;
};

struct vertex* vertex_malloc(const char* name, int number);
struct vertex* vertex_realloc(char* name, int number);
int vertex_cmp(struct vertex* v1, struct vertex* v2);
void vertex_print(struct vertex* v);
void vertex_delete(struct vertex* v);

struct edge {
    int vertex1;
    int vertex2;
    float t_start;
    float t_end;
};

struct edge* edge_malloc(int vertex1, int vertex2, float t_start, float t_end);
struct edge* edge_realloc(int vertex1, int vertex2);
int edge_cmp(struct edge* e1, struct edge* e2);
void edge_print(struct edge* e);
void edge_delete(struct edge* e);

struct graph {
    struct vertex** vertex_set_by_name;
    struct vertex** vertex_set_by_number;
    int n_vertices;
    int n_edges;
    struct edge** edge_set; // temp array for edge reading
    struct spmat sp; // final edge set as sparse matrix yale format
};

struct graph* graph_malloc(void);
void graph_vertex_malloc(struct graph* g, int n_vertices);
void graph_edge_malloc(struct graph* g, int n_edges);

struct graph* graph_read(char* file_name);
void graph_edge_list_read(struct graph* g, FILE*input_file, char* file_name);
void graph_read_vertices(struct graph* g, FILE* input_file, char* file_name);
void graph_read_edges(struct graph* g, FILE* input_file, char* file_name);
void graph_sif_read(struct graph *g, FILE *input_file, char *file_name);

void graph_print_events(struct graph* g);

void graph_print_vertices(struct graph* g);
void graph_print_edges(struct graph* g);
void graph_sort_vertices(struct graph* g);
void graph_sort_edges(struct graph* g);
int graph_find_vertex_number(struct graph* g, char* name);
char* graph_find_vertex_name(struct graph* g, int number);
int graph_find_edge(struct graph* g, int v1, int v2);
int graph_is_edge(struct graph* g, int vertex1, int vertex2);
void graph_delete(struct graph* g);

struct alignment {
    struct graph* domain;
    struct graph* range;
    struct permutation* perm;
    //	struct permutation* invperm;
    int n_edges_preserved;
    int n_edges_induced;
    int n_edges_unique;
    float t_duration_conserved;
    float t_duration_nonconserved;
    float score, edge_score, node_score;
    int is_computed;
    int flipgorder;
};

struct dymatelt {
    int n; // number of events
    int ne; // number of edges
    int alloc; // number of allocated events
    float *start;
    float *end;
};

struct compmatelt {
    int lr;
    int nl; // number of events
    int nr; // number of events
    int ne; // number of edges
    float *startl;
    float *endl;
    float *startr;
    float *endr;
};
void compmatelts_print(struct compmatelt* spA, int m);

struct compute_aux_space {
//    SpMat cg; // composite graphs
//    std::vector<EigTriplet> cg_triplets;
//    int *v_weights;
    int nnz;
    int *spIA; // ind first elt of each row
    int *spJA; // col inds
    int *spJR; // col inds
    int *spJC; // row inds
    int *spINZ; // nnz of each row
    int *spWI; // for the accum. could be done w/ spINZ
    int *v_weights;
    struct compmatelt *tempA;
    struct compmatelt *spA;
    float *tempstart;
    float *tempend;
};

void transfer_event_values(struct dymatelt *spA,
                           int count, int shift, int k, float *tempstart, float *tempend);
void transfer_event_values_sp(struct spmat spran, struct dymatelt *spA,
                              int u, int i);


struct alignment* alignment_calloc(struct graph* g1, struct graph* g2);
void alignment_randomize(struct alignment* a);
int alignment_read(struct alignment* a, char* file_name, int flipgorder);
void alignment_tensor(struct alignment* a3, struct alignment* a1, struct alignment* a2, struct tensor_aux_space* taux);
int alignment_is_preserved(struct alignment* a, struct edge* e, struct edge* e_temp);
int alignment_is_mapped(struct alignment* a, struct edge* e, struct edge* e_temp);
int alignment_edge_correctness(struct alignment* a);
int alignment_induced_edge_correctness(struct alignment* a);
int alignment_compare(struct alignment* a1, struct alignment* a2, struct carrier* rel);
int alignment_compare_ec(const void* a1, const void* a2);
int alignment_compare_ics(const void* a1, const void* a2);
int alignment_compare_3s(const void* a1, const void* a2);
void alignment_update_inverse(struct alignment* a);
void alignment_write(struct alignment* a, char* file_name);
void alignment_graph_write(struct alignment* a, char* file_name);
void alignment_common_subgraph_write(struct alignment* a, char* file_name,
                                     compute_aux_space *caux);
void alignment_print(struct alignment* a);
void alignment_delete(struct alignment* a);
float alignment_nodescore_compute(struct alignment *a, struct carrier *rel);
float alignment_node_correctness(struct alignment *a);

// float alignment_partial_substructure_score(SpMat G, int *v_weights);
// void alignment_composite_graph(struct alignment *a, compute_aux_space *caux);
// void alignment_edges_preserved(SpMat& G, int *v_weights,
//                                struct alignment *a,
//                                int *n_edges_preserved, int *n_unique_edges);
void alignment_compute(struct alignment* a, struct carrier* rel,
                       compute_aux_space *caux);

struct compute_aux_space* compute_aux_space_malloc(struct graph *g1, struct graph *g2);
void compute_aux_space_delete(struct compute_aux_space* aux);


#endif
